@@ -7,7 +7,7 @@ package fs
 import (
 	"context"
 	"log"
-	"math/rand"
+	"runtime/debug"
 	"sync"
 	"syscall"
 	"time"
@@ -30,9 +30,17 @@ type fileEntry struct {
 	mu sync.Mutex
 
 	// Directory
-	dirStream   DirStream
-	hasOverflow bool
-	overflow    fuse.DirEntry
+	dirStream     DirStream
+	hasOverflow   bool
+	overflow      fuse.DirEntry
+	overflowErrno syscall.Errno
+
+	// dirOffset is the current location in the directory (see `telldir(3)`).
+	// The value is equivalent to `d_off` (see `getdents(2)`) of the last
+	// directory entry sent to the kernel so far.
+	// If `dirOffset` and `fuse.DirEntryList.offset` disagree, then a
+	// directory seek has taken place.
+	dirOffset uint64
 
 	wg sync.WaitGroup
 }
@@ -55,10 +63,34 @@ type rawBridge struct {
 
 	// mu protects the following data.  Locks for inodes must be
 	// taken before rawBridge.mu
-	mu           sync.Mutex
-	nodes        map[uint64]*Inode
+	mu sync.Mutex
+
+	// stableAttrs is used to detect already-known nodes and hard links by
+	// looking at:
+	// 1) file type ......... StableAttr.Mode
+	// 2) inode number ...... StableAttr.Ino
+	// 3) generation number . StableAttr.Gen
+	stableAttrs  map[StableAttr]*Inode
 	automaticIno uint64
 
+	// The *Node ID* is an arbitrary uint64 identifier chosen by the FUSE library.
+	// It is used the identify *nodes* (files/directories/symlinks/...) in the
+	// communication between the FUSE library and the Linux kernel.
+	//
+	// The kernelNodeIds map translates between the NodeID and the corresponding
+	// go-fuse Inode object.
+	//
+	// A simple incrementing counter is used as the NodeID (see `nextNodeID`).
+	kernelNodeIds map[uint64]*Inode
+	// nextNodeID is the next free NodeID. Increment after copying the value.
+	nextNodeId uint64
+	// nodeCountHigh records the highest number of entries we had in the
+	// kernelNodeIds map.
+	// As the size of stableAttrs tracks kernelNodeIds (+- a few entries due to
+	// concurrent FORGETs, LOOKUPs, and the fixed NodeID 1), this is also a good
+	// estimate for stableAttrs.
+	nodeCountHigh int
+
 	files     []*fileEntry
 	freeFiles []uint32
 }
@@ -77,55 +109,26 @@ func (b *rawBridge) newInodeUnlocked(ops InodeEmbedder, id StableAttr, persisten
 		return ops.embed()
 	}
 
-	if id.Ino == 0 {
-		for {
-			id.Ino = b.automaticIno
-			b.automaticIno++
-			_, ok := b.nodes[id.Ino]
-			if !ok {
-				break
-			}
-		}
-	}
-
 	// Only the file type bits matter
 	id.Mode = id.Mode & syscall.S_IFMT
 	if id.Mode == 0 {
 		id.Mode = fuse.S_IFREG
 	}
 
-	// the same node can be looked up through 2 paths in parallel, eg.
-	//
-	//	    root
-	//	    /  \
-	//	  dir1 dir2
-	//	    \  /
-	//	    file
-	//
-	// dir1.Lookup("file") and dir2.Lookup("file") are executed
-	// simultaneously.  The matching StableAttrs ensure that we return the
-	// same node.
-	var t time.Duration
-	t0 := time.Now()
-	for i := 1; true; i++ {
-		old := b.nodes[id.Ino]
-		if old == nil {
-			break
+	if id.Ino == 0 {
+		// Find free inode number.
+		for {
+			id.Ino = b.automaticIno
+			b.automaticIno++
+			_, ok := b.stableAttrs[id]
+			if !ok {
+				break
+			}
 		}
-		if old.stableAttr == id {
-			return old
-		}
-		b.mu.Unlock()
-
-		t = expSleep(t)
-		if i%5000 == 0 {
-			b.logf("blocked for %.0f seconds waiting for FORGET on i%d", time.Since(t0).Seconds(), id.Ino)
-		}
-		b.mu.Lock()
 	}
 
-	b.nodes[id.Ino] = ops.embed()
-	initInode(ops.embed(), ops, id, b, persistent)
+	initInode(ops.embed(), ops, id, b, persistent, b.nextNodeId)
+	b.nextNodeId++
 	return ops.embed()
 }
 
@@ -135,21 +138,6 @@ func (b *rawBridge) logf(format string, args ...interface{}) {
 	}
 }
 
-// expSleep sleeps for time `t` and returns an exponentially increasing value
-// for the next sleep time, capped at 1 ms.
-func expSleep(t time.Duration) time.Duration {
-	if t == 0 {
-		return time.Microsecond
-	}
-	time.Sleep(t)
-	// Next sleep is between t and 2*t
-	t += time.Duration(rand.Int63n(int64(t)))
-	if t >= time.Millisecond {
-		return time.Millisecond
-	}
-	return t
-}
-
 func (b *rawBridge) newInode(ctx context.Context, ops InodeEmbedder, id StableAttr, persistent bool) *Inode {
 	ch := b.newInodeUnlocked(ops, id, persistent)
 	if ch != ops.embed() {
@@ -163,34 +151,84 @@ func (b *rawBridge) newInode(ctx context.Context, ops InodeEmbedder, id StableAt
 }
 
 // addNewChild inserts the child into the tree. Returns file handle if file != nil.
-func (b *rawBridge) addNewChild(parent *Inode, name string, child *Inode, file FileHandle, fileFlags uint32, out *fuse.EntryOut) uint32 {
+// Unless fileFlags has the syscall.O_EXCL bit set, child.stableAttr will be used
+// to find an already-known node. If one is found, `child` is ignored and the
+// already-known one is used. The node that was actually used is returned.
+func (b *rawBridge) addNewChild(parent *Inode, name string, child *Inode, file FileHandle, fileFlags uint32, out *fuse.EntryOut) (selected *Inode, fh uint32) {
 	if name == "." || name == ".." {
 		log.Panicf("BUG: tried to add virtual entry %q to the actual tree", name)
 	}
-	lockNodes(parent, child)
-	parent.setEntry(name, child)
-	b.mu.Lock()
 
-	// Due to concurrent FORGETs, lookupCount may have dropped to zero.
-	// This means it MAY have been deleted from nodes[] already. Add it back.
-	if child.lookupCount == 0 {
-		b.nodes[child.stableAttr.Ino] = child
+	// the same node can be looked up through 2 paths in parallel, eg.
+	//
+	//	    root
+	//	    /  \
+	//	  dir1 dir2
+	//	    \  /
+	//	    file
+	//
+	// dir1.Lookup("file") and dir2.Lookup("file") are executed
+	// simultaneously.  The matching StableAttrs ensure that we return the
+	// same node.
+	orig := child
+	id := child.stableAttr
+	if id.Mode & ^(uint32(syscall.S_IFMT)) != 0 {
+		log.Panicf("%#v", id)
 	}
+	for {
+		lockNodes(parent, child)
+		b.mu.Lock()
+		if fileFlags&syscall.O_EXCL != 0 {
+			// must create a new node - don't look for existing nodes
+			break
+		}
+		old := b.stableAttrs[id]
+		if old == nil {
+			if child == orig {
+				// no pre-existing node under this inode number
+				break
+			} else {
+				// old inode disappeared while we were looping here. Go back to
+				// original child.
+				b.mu.Unlock()
+				unlockNodes(parent, child)
+				child = orig
+				continue
+			}
+		}
+		if old == child {
+			// we now have the right inode locked
+			break
+		}
+		// found a different existing node
+		b.mu.Unlock()
+		unlockNodes(parent, child)
+		child = old
+	}
+
 	child.lookupCount++
 	child.changeCounter++
 
-	var fh uint32
+	b.kernelNodeIds[child.nodeId] = child
+	if len(b.kernelNodeIds) > b.nodeCountHigh {
+		b.nodeCountHigh = len(b.kernelNodeIds)
+	}
+	// Any node that might be there is overwritten - it is obsolete now
+	b.stableAttrs[id] = child
 	if file != nil {
 		fh = b.registerFile(child, file, fileFlags)
 	}
 
-	out.NodeId = child.stableAttr.Ino
+	parent.setEntry(name, child)
+
+	out.NodeId = child.nodeId
 	out.Generation = child.stableAttr.Gen
 	out.Attr.Ino = child.stableAttr.Ino
 
 	b.mu.Unlock()
 	unlockNodes(parent, child)
-	return fh
+
+	return child, fh
 }
 
 func (b *rawBridge) setEntryOutTimeout(out *fuse.EntryOut) {
@@ -231,9 +269,8 @@ func NewNodeFS(root InodeEmbedder, opts *Options) fuse.RawFileSystem {
 	bridge := &rawBridge{
 		automaticIno: opts.FirstAutomaticIno,
 		server:       opts.ServerCallbacks,
-	}
-	if bridge.automaticIno == 1 {
-		bridge.automaticIno++
+		nextNodeId:   2, // the root node has nodeid 1
+		stableAttrs:  make(map[StableAttr]*Inode),
 	}
 
 	if bridge.automaticIno == 0 {
@@ -248,17 +285,24 @@ func NewNodeFS(root InodeEmbedder, opts *Options) fuse.RawFileSystem {
 		bridge.options.AttrTimeout = &oneSec
 	}
 
+	stableAttr := StableAttr{
+		Ino:  root.embed().StableAttr().Ino,
+		Mode: fuse.S_IFDIR,
+	}
+	if opts.RootStableAttr != nil {
+		stableAttr.Ino = opts.RootStableAttr.Ino
+		stableAttr.Gen = opts.RootStableAttr.Gen
+	}
+
 	initInode(root.embed(), root,
-		StableAttr{
-			Ino:  1,
-			Mode: fuse.S_IFDIR,
-		},
+		stableAttr,
 		bridge,
 		false,
+		1,
 	)
 	bridge.root = root.embed()
 	bridge.root.lookupCount = 1
-	bridge.nodes = map[uint64]*Inode{
+	bridge.kernelNodeIds = map[uint64]*Inode{
 		1: bridge.root,
 	}
 
@@ -281,7 +325,7 @@ func (b *rawBridge) String() string {
 func (b *rawBridge) inode(id uint64, fh uint64) (*Inode, *fileEntry) {
 	b.mu.Lock()
 	defer b.mu.Unlock()
-	n, f := b.nodes[id], b.files[fh]
+	n, f := b.kernelNodeIds[id], b.files[fh]
 	if n == nil {
 		log.Panicf("unknown node %d", id)
 	}
@@ -300,8 +344,8 @@ func (b *rawBridge) Lookup(cancel <-chan struct{}, header *fuse.InHeader, name s
 		return errnoToStatus(errno)
 	}
 
+	child, _ = b.addNewChild(parent, name, child, nil, 0, out)
 	child.setEntryOut(out)
-	b.addNewChild(parent, name, child, nil, 0, out)
 	b.setEntryOutTimeout(out)
 	return fuse.OK
 }
@@ -334,6 +378,8 @@ func (b *rawBridge) Rmdir(cancel <-chan struct{}, header *fuse.InHeader, name st
 		errno = mops.Rmdir(&fuse.Context{Caller: header.Caller, Cancel: cancel}, name)
 	}
 
+	// TODO - this should not succeed silently.
+
 	if errno == 0 {
 		parent.RmChild(name)
 	}
@@ -347,6 +393,8 @@ func (b *rawBridge) Unlink(cancel <-chan struct{}, header *fuse.InHeader, name s
 		errno = mops.Unlink(&fuse.Context{Caller: header.Caller, Cancel: cancel}, name)
 	}
 
+	// TODO - this should not succeed silently.
+
 	if errno == 0 {
 		parent.RmChild(name)
 	}
@@ -376,8 +424,8 @@ func (b *rawBridge) Mkdir(cancel <-chan struct{}, input *fuse.MkdirIn, name stri
 		log.Panicf("Mkdir: mode must be S_IFDIR (%o), got %o", fuse.S_IFDIR, out.Attr.Mode)
 	}
 
+	child, _ = b.addNewChild(parent, name, child, nil, syscall.O_EXCL, out)
 	child.setEntryOut(out)
-	b.addNewChild(parent, name, child, nil, 0, out)
 	b.setEntryOutTimeout(out)
 	return fuse.OK
 }
@@ -389,14 +437,16 @@ func (b *rawBridge) Mknod(cancel <-chan struct{}, input *fuse.MknodIn, name stri
 	var errno syscall.Errno
 	if mops, ok := parent.ops.(NodeMknoder); ok {
 		child, errno = mops.Mknod(&fuse.Context{Caller: input.Caller, Cancel: cancel}, name, input.Mode, input.Rdev, out)
+	} else {
+		return fuse.ENOTSUP
 	}
 
 	if errno != 0 {
 		return errnoToStatus(errno)
 	}
 
+	child, _ = b.addNewChild(parent, name, child, nil, syscall.O_EXCL, out)
 	child.setEntryOut(out)
-	b.addNewChild(parent, name, child, nil, 0, out)
 	b.setEntryOutTimeout(out)
 	return fuse.OK
 }
@@ -422,8 +472,9 @@ func (b *rawBridge) Create(cancel <-chan struct{}, input *fuse.CreateIn, name st
 		return errnoToStatus(errno)
 	}
 
-	out.Fh = uint64(b.addNewChild(parent, name, child, f, input.Flags|syscall.O_CREAT, &out.EntryOut))
+	child, fh := b.addNewChild(parent, name, child, f, input.Flags|syscall.O_CREAT|syscall.O_EXCL, &out.EntryOut)
 
+	out.Fh = uint64(fh)
 	out.OpenFlags = flags
 
 	child.setEntryOut(&out.EntryOut)
@@ -433,7 +484,47 @@ func (b *rawBridge) Create(cancel <-chan struct{}, input *fuse.CreateIn, name st
 
 func (b *rawBridge) Forget(nodeid, nlookup uint64) {
 	n, _ := b.inode(nodeid, 0)
-	n.removeRef(nlookup, false)
+	forgotten, _ := n.removeRef(nlookup, false)
+
+	if forgotten {
+		b.compactMemory()
+	}
+}
+
+// compactMemory tries to free memory that was previously used by forgotten
+// nodes.
+//
+// Maps do not free all memory when elements get deleted
+// ( https://github.com/golang/go/issues/20135 ).
+// As a workaround, we recreate our two big maps (stableAttrs & kernelNodeIds)
+// every time they have shrunk dramatically (100 x smaller).
+// In this case, `nodeCountHigh` is reset to the new (smaller) size.
+func (b *rawBridge) compactMemory() {
+	b.mu.Lock()
+
+	if b.nodeCountHigh <= len(b.kernelNodeIds)*100 {
+		b.mu.Unlock()
+		return
+	}
+
+	tmpStableAttrs := make(map[StableAttr]*Inode, len(b.stableAttrs))
+	for i, v := range b.stableAttrs {
+		tmpStableAttrs[i] = v
+	}
+	b.stableAttrs = tmpStableAttrs
+
+	tmpKernelNodeIds := make(map[uint64]*Inode, len(b.kernelNodeIds))
+	for i, v := range b.kernelNodeIds {
+		tmpKernelNodeIds[i] = v
+	}
+	b.kernelNodeIds = tmpKernelNodeIds
+
+	b.nodeCountHigh = len(b.kernelNodeIds)
+
+	b.mu.Unlock()
+
+	// Run outside b.mu
+	debug.FreeOSMemory()
 }
 
 func (b *rawBridge) SetDebug(debug bool) {}
@@ -461,15 +552,10 @@ func (b *rawBridge) GetAttr(cancel <-chan struct{}, input *fuse.GetAttrIn, out *
 func (b *rawBridge) getattr(ctx context.Context, n *Inode, f FileHandle, out *fuse.AttrOut) syscall.Errno {
 	var errno syscall.Errno
 
-	var fg FileGetattrer
-	if f != nil {
-		fg, _ = f.(FileGetattrer)
-	}
-
-	if fops, ok := n.ops.(NodeGetattrer); ok {
-		errno = fops.Getattr(ctx, f, out)
-	} else if fg != nil {
-		errno = fg.Getattr(ctx, out)
+	if nodeOps, ok := n.ops.(NodeGetattrer); ok {
+		errno = nodeOps.Getattr(ctx, f, out)
+	} else if fileOps, ok := f.(FileGetattrer); ok {
+		errno = fileOps.Getattr(ctx, out)
 	} else {
 		// We set Mode below, which is the minimum for success
 	}
@@ -534,8 +620,8 @@ func (b *rawBridge) Link(cancel <-chan struct{}, input *fuse.LinkIn, name string
 			return errnoToStatus(errno)
 		}
 
+		child, _ = b.addNewChild(parent, name, child, nil, 0, out)
 		child.setEntryOut(out)
-		b.addNewChild(parent, name, child, nil, 0, out)
 		b.setEntryOutTimeout(out)
 		return fuse.OK
 	}
@@ -551,7 +637,7 @@ func (b *rawBridge) Symlink(cancel <-chan struct{}, header *fuse.InHeader, targe
 			return errnoToStatus(status)
 		}
 
-		b.addNewChild(parent, name, child, nil, 0, out)
+		child, _ = b.addNewChild(parent, name, child, nil, syscall.O_EXCL, out)
 		child.setEntryOut(out)
 		b.setEntryOutTimeout(out)
 		return fuse.OK
@@ -679,12 +765,13 @@ func (b *rawBridge) registerFile(n *Inode, f FileHandle, flags uint32) uint32 {
 func (b *rawBridge) Read(cancel <-chan struct{}, input *fuse.ReadIn, buf []byte) (fuse.ReadResult, fuse.Status) {
 	n, f := b.inode(input.NodeId, input.Fh)
 
+	ctx := &fuse.Context{Caller: input.Caller, Cancel: cancel}
 	if fops, ok := n.ops.(NodeReader); ok {
-		res, errno := fops.Read(&fuse.Context{Caller: input.Caller, Cancel: cancel}, f.file, buf, int64(input.Offset))
+		res, errno := fops.Read(ctx, f.file, buf, int64(input.Offset))
 		return res, errnoToStatus(errno)
 	}
 	if fr, ok := f.file.(FileReader); ok {
-		res, errno := fr.Read(&fuse.Context{Caller: input.Caller, Cancel: cancel}, buf, int64(input.Offset))
+		res, errno := fr.Read(ctx, buf, int64(input.Offset))
 		return res, errnoToStatus(errno)
 	}
 
@@ -694,32 +781,35 @@ func (b *rawBridge) Read(cancel <-chan struct{}, input *fuse.ReadIn, buf []byte)
 func (b *rawBridge) GetLk(cancel <-chan struct{}, input *fuse.LkIn, out *fuse.LkOut) fuse.Status {
 	n, f := b.inode(input.NodeId, input.Fh)
 
+	ctx := &fuse.Context{Caller: input.Caller, Cancel: cancel}
 	if lops, ok := n.ops.(NodeGetlker); ok {
-		return errnoToStatus(lops.Getlk(&fuse.Context{Caller: input.Caller, Cancel: cancel}, f.file, input.Owner, &input.Lk, input.LkFlags, &out.Lk))
+		return errnoToStatus(lops.Getlk(ctx, f.file, input.Owner, &input.Lk, input.LkFlags, &out.Lk))
 	}
 	if gl, ok := f.file.(FileGetlker); ok {
-		return errnoToStatus(gl.Getlk(&fuse.Context{Caller: input.Caller, Cancel: cancel}, input.Owner, &input.Lk, input.LkFlags, &out.Lk))
+		return errnoToStatus(gl.Getlk(ctx, input.Owner, &input.Lk, input.LkFlags, &out.Lk))
 	}
 	return fuse.ENOTSUP
 }
 
 func (b *rawBridge) SetLk(cancel <-chan struct{}, input *fuse.LkIn) fuse.Status {
 	n, f := b.inode(input.NodeId, input.Fh)
+	ctx := &fuse.Context{Caller: input.Caller, Cancel: cancel}
 	if lops, ok := n.ops.(NodeSetlker); ok {
-		return errnoToStatus(lops.Setlk(&fuse.Context{Caller: input.Caller, Cancel: cancel}, f.file, input.Owner, &input.Lk, input.LkFlags))
+		return errnoToStatus(lops.Setlk(ctx, f.file, input.Owner, &input.Lk, input.LkFlags))
 	}
-	if sl, ok := n.ops.(FileSetlker); ok {
-		return errnoToStatus(sl.Setlk(&fuse.Context{Caller: input.Caller, Cancel: cancel}, input.Owner, &input.Lk, input.LkFlags))
+	if sl, ok := f.file.(FileSetlker); ok {
+		return errnoToStatus(sl.Setlk(ctx, input.Owner, &input.Lk, input.LkFlags))
 	}
 	return fuse.ENOTSUP
 }
 func (b *rawBridge) SetLkw(cancel <-chan struct{}, input *fuse.LkIn) fuse.Status {
 	n, f := b.inode(input.NodeId, input.Fh)
+	ctx := &fuse.Context{Caller: input.Caller, Cancel: cancel}
 	if lops, ok := n.ops.(NodeSetlkwer); ok {
-		return errnoToStatus(lops.Setlkw(&fuse.Context{Caller: input.Caller, Cancel: cancel}, f.file, input.Owner, &input.Lk, input.LkFlags))
+		return errnoToStatus(lops.Setlkw(ctx, f.file, input.Owner, &input.Lk, input.LkFlags))
 	}
-	if sl, ok := n.ops.(FileSetlkwer); ok {
-		return errnoToStatus(sl.Setlkw(&fuse.Context{Caller: input.Caller, Cancel: cancel}, input.Owner, &input.Lk, input.LkFlags))
+	if sl, ok := f.file.(FileSetlkwer); ok {
+		return errnoToStatus(sl.Setlkw(ctx, input.Owner, &input.Lk, input.LkFlags))
 	}
 	return fuse.ENOTSUP
 }
@@ -732,10 +822,11 @@ func (b *rawBridge) Release(cancel <-chan struct{}, input *fuse.ReleaseIn) {
 
 	f.wg.Wait()
 
+	ctx := &fuse.Context{Caller: input.Caller, Cancel: cancel}
 	if r, ok := n.ops.(NodeReleaser); ok {
-		r.Release(&fuse.Context{Caller: input.Caller, Cancel: cancel}, f.file)
+		r.Release(ctx, f.file)
 	} else if r, ok := f.file.(FileReleaser); ok {
-		r.Release(&fuse.Context{Caller: input.Caller, Cancel: cancel})
+		r.Release(ctx)
 	}
 
 	b.mu.Lock()
@@ -763,7 +854,7 @@ func (b *rawBridge) releaseFileEntry(nid uint64, fh uint64) (*Inode, *fileEntry)
 	b.mu.Lock()
 	defer b.mu.Unlock()
 
-	n := b.nodes[nid]
+	n := b.kernelNodeIds[nid]
 	var entry *fileEntry
 	if fh > 0 {
 		last := len(n.openFiles) - 1
@@ -781,12 +872,13 @@ func (b *rawBridge) releaseFileEntry(nid uint64, fh uint64) (*Inode, *fileEntry)
 func (b *rawBridge) Write(cancel <-chan struct{}, input *fuse.WriteIn, data []byte) (written uint32, status fuse.Status) {
 	n, f := b.inode(input.NodeId, input.Fh)
 
+	ctx := &fuse.Context{Caller: input.Caller, Cancel: cancel}
 	if wr, ok := n.ops.(NodeWriter); ok {
-		w, errno := wr.Write(&fuse.Context{Caller: input.Caller, Cancel: cancel}, f.file, data, int64(input.Offset))
+		w, errno := wr.Write(ctx, f.file, data, int64(input.Offset))
 		return w, errnoToStatus(errno)
 	}
 	if fr, ok := f.file.(FileWriter); ok {
-		w, errno := fr.Write(&fuse.Context{Caller: input.Caller, Cancel: cancel}, data, int64(input.Offset))
+		w, errno := fr.Write(ctx, data, int64(input.Offset))
 		return w, errnoToStatus(errno)
 	}
 
@@ -795,33 +887,36 @@ func (b *rawBridge) Write(cancel <-chan struct{}, input *fuse.WriteIn, data []by
 
 func (b *rawBridge) Flush(cancel <-chan struct{}, input *fuse.FlushIn) fuse.Status {
 	n, f := b.inode(input.NodeId, input.Fh)
+	ctx := &fuse.Context{Caller: input.Caller, Cancel: cancel}
 	if fl, ok := n.ops.(NodeFlusher); ok {
-		return errnoToStatus(fl.Flush(&fuse.Context{Caller: input.Caller, Cancel: cancel}, f.file))
+		return errnoToStatus(fl.Flush(ctx, f.file))
 	}
 	if fl, ok := f.file.(FileFlusher); ok {
-		return errnoToStatus(fl.Flush(&fuse.Context{Caller: input.Caller, Cancel: cancel}))
+		return errnoToStatus(fl.Flush(ctx))
 	}
 	return 0
 }
 
 func (b *rawBridge) Fsync(cancel <-chan struct{}, input *fuse.FsyncIn) fuse.Status {
 	n, f := b.inode(input.NodeId, input.Fh)
+	ctx := &fuse.Context{Caller: input.Caller, Cancel: cancel}
 	if fs, ok := n.ops.(NodeFsyncer); ok {
-		return errnoToStatus(fs.Fsync(&fuse.Context{Caller: input.Caller, Cancel: cancel}, f.file, input.FsyncFlags))
+		return errnoToStatus(fs.Fsync(ctx, f.file, input.FsyncFlags))
 	}
 	if fs, ok := f.file.(FileFsyncer); ok {
-		return errnoToStatus(fs.Fsync(&fuse.Context{Caller: input.Caller, Cancel: cancel}, input.FsyncFlags))
+		return errnoToStatus(fs.Fsync(ctx, input.FsyncFlags))
 	}
 	return fuse.ENOTSUP
 }
 
 func (b *rawBridge) Fallocate(cancel <-chan struct{}, input *fuse.FallocateIn) fuse.Status {
 	n, f := b.inode(input.NodeId, input.Fh)
+	ctx := &fuse.Context{Caller: input.Caller, Cancel: cancel}
 	if a, ok := n.ops.(NodeAllocater); ok {
-		return errnoToStatus(a.Allocate(&fuse.Context{Caller: input.Caller, Cancel: cancel}, f.file, input.Offset, input.Length, input.Mode))
+		return errnoToStatus(a.Allocate(ctx, f.file, input.Offset, input.Length, input.Mode))
 	}
 	if a, ok := f.file.(FileAllocater); ok {
-		return errnoToStatus(a.Allocate(&fuse.Context{Caller: input.Caller, Cancel: cancel}, input.Offset, input.Length, input.Mode))
+		return errnoToStatus(a.Allocate(ctx, input.Offset, input.Length, input.Mode))
 	}
 	return fuse.ENOTSUP
 }
@@ -842,23 +937,47 @@ func (b *rawBridge) OpenDir(cancel <-chan struct{}, input *fuse.OpenIn, out *fus
 	return fuse.OK
 }
 
-// setStream sets the directory part of f. Must hold f.mu
-func (b *rawBridge) setStream(cancel <-chan struct{}, input *fuse.ReadIn, inode *Inode, f *fileEntry) syscall.Errno {
-	if f.dirStream == nil || input.Offset == 0 {
+// setStream makes sure `f.dirStream` and associated state variables are set and
+// seeks to offset requested in `input`. Caller must hold `f.mu`.
+// The `eof` return value shows if `f.dirStream` ended before the requested
+// offset was reached.
+func (b *rawBridge) setStream(cancel <-chan struct{}, input *fuse.ReadIn, inode *Inode, f *fileEntry) (errno syscall.Errno, eof bool) {
+	// Get a new directory stream in the following cases:
+	// 1) f.dirStream == nil ............ First READDIR[PLUS] on this file handle.
+	// 2) input.Offset == 0 ............. Start reading the directory again from
+	//                                    the beginning (user called rewinddir(3) or lseek(2)).
+	// 3) input.Offset < f.nextOffset ... Seek back (user called seekdir(3) or lseek(2)).
+	if f.dirStream == nil || input.Offset == 0 || input.Offset < f.dirOffset {
 		if f.dirStream != nil {
 			f.dirStream.Close()
 			f.dirStream = nil
 		}
 		str, errno := b.getStream(&fuse.Context{Caller: input.Caller, Cancel: cancel}, inode)
 		if errno != 0 {
-			return errno
+			return errno, false
 		}
 
+		f.dirOffset = 0
 		f.hasOverflow = false
 		f.dirStream = str
 	}
 
-	return 0
+	// Seek forward?
+	for f.dirOffset < input.Offset {
+		f.hasOverflow = false
+		if !f.dirStream.HasNext() {
+			// Seek past end of directory. This is not an error, but the
+			// user will get an empty directory listing.
+			return 0, true
+		}
+		_, errno := f.dirStream.Next()
+		if errno != 0 {
+			return errno, true
+		}
+		f.dirOffset++
+	}
+
+	return 0, false
 }
 
 func (b *rawBridge) getStream(ctx context.Context, inode *Inode) (DirStream, syscall.Errno) {
@@ -866,11 +985,12 @@ func (b *rawBridge) getStream(ctx context.Context, inode *Inode) (DirStream, sys
 		return rd.Readdir(ctx)
 	}
 
-	r := []fuse.DirEntry{}
-	for k, ch := range inode.Children() {
-		r = append(r, fuse.DirEntry{Mode: ch.Mode(),
-			Name: k,
-			Ino:  ch.StableAttr().Ino})
+	lst := inode.childrenList()
+	r := make([]fuse.DirEntry, 0, len(lst))
+	for _, e := range lst {
+		r = append(r, fuse.DirEntry{Mode: e.Inode.Mode(),
+			Name: e.Name,
+			Ino:  e.Inode.StableAttr().Ino})
 	}
 	return NewListDirStream(r), 0
 }
@@ -880,27 +1000,46 @@ func (b *rawBridge) ReadDir(cancel <-chan struct{}, input *fuse.ReadIn, out *fus
 
 	f.mu.Lock()
 	defer f.mu.Unlock()
-	if errno := b.setStream(cancel, input, n, f); errno != 0 {
+
+	errno, eof := b.setStream(cancel, input, n, f)
+	if errno != 0 {
 		return errnoToStatus(errno)
+	} else if eof {
+		return fuse.OK
 	}
 
 	if f.hasOverflow {
+		if f.overflowErrno != 0 {
+			return errnoToStatus(f.overflowErrno)
+		}
+
+		f.hasOverflow = false
 		// always succeeds.
 		out.AddDirEntry(f.overflow)
-		f.hasOverflow = false
+		f.dirOffset++
 	}
 
+	first := true
 	for f.dirStream.HasNext() {
 		e, errno := f.dirStream.Next()
 
 		if errno != 0 {
-			return errnoToStatus(errno)
+			if first {
+				return errnoToStatus(errno)
+			} else {
+				f.hasOverflow = true
+				f.overflowErrno = errno
+				return fuse.OK
+			}
 		}
+
+		first = false
 		if !out.AddDirEntry(e) {
 			f.overflow = e
 			f.hasOverflow = true
 			return errnoToStatus(errno)
 		}
+		f.dirOffset++
 	}
 
 	return fuse.OK
@@ -911,25 +1050,39 @@ func (b *rawBridge) ReadDirPlus(cancel <-chan struct{}, input *fuse.ReadIn, out
 
 	f.mu.Lock()
 	defer f.mu.Unlock()
-	if errno := b.setStream(cancel, input, n, f); errno != 0 {
+
+	errno, eof := b.setStream(cancel, input, n, f)
+	if errno != 0 {
 		return errnoToStatus(errno)
+	} else if eof {
+		return fuse.OK
 	}
 
 	ctx := &fuse.Context{Caller: input.Caller, Cancel: cancel}
+
+	first := true
 	for f.dirStream.HasNext() || f.hasOverflow {
 		var e fuse.DirEntry
 		var errno syscall.Errno
 
 		if f.hasOverflow {
 			e = f.overflow
+			errno = f.overflowErrno
 			f.hasOverflow = false
 		} else {
 			e, errno = f.dirStream.Next()
 		}
 
 		if errno != 0 {
-			return errnoToStatus(errno)
+			if first {
+				return errnoToStatus(errno)
+			} else {
+				f.overflowErrno = errno
+				f.hasOverflow = true
+				return fuse.OK
+			}
 		}
+		first = false
 
 		entryOut := out.AddDirLookupEntry(e)
 		if entryOut == nil {
@@ -937,6 +1090,7 @@ func (b *rawBridge) ReadDirPlus(cancel <-chan struct{}, input *fuse.ReadIn, out
 			f.hasOverflow = true
 			return fuse.OK
 		}
+		f.dirOffset++
 
 		// Virtual entries "." and ".." should be part of the
 		// directory listing, but not part of the filesystem tree.
@@ -953,7 +1107,7 @@ func (b *rawBridge) ReadDirPlus(cancel <-chan struct{}, input *fuse.ReadIn, out
 				entryOut.SetEntryTimeout(*b.options.NegativeTimeout)
 			}
 		} else {
-			b.addNewChild(n, e.Name, child, nil, 0, entryOut)
+			child, _ = b.addNewChild(n, e.Name, child, nil, 0, entryOut)
 			child.setEntryOut(entryOut)
 			b.setEntryOutTimeout(entryOut)
 			if e.Mode&syscall.S_IFMT != child.stableAttr.Mode&syscall.S_IFMT {
@@ -1007,23 +1161,39 @@ func (b *rawBridge) CopyFileRange(cancel <-chan struct{}, in *fuse.CopyFileRange
 func (b *rawBridge) Lseek(cancel <-chan struct{}, in *fuse.LseekIn, out *fuse.LseekOut) fuse.Status {
 	n, f := b.inode(in.NodeId, in.Fh)
 
+	ctx := &fuse.Context{Caller: in.Caller, Cancel: cancel}
+
 	ls, ok := n.ops.(NodeLseeker)
 	if ok {
-		off, errno := ls.Lseek(&fuse.Context{Caller: in.Caller, Cancel: cancel},
+		off, errno := ls.Lseek(ctx,
 			f.file, in.Offset, in.Whence)
 		out.Offset = off
 		return errnoToStatus(errno)
 	}
 	if fs, ok := f.file.(FileLseeker); ok {
-		off, errno := fs.Lseek(&fuse.Context{Caller: in.Caller, Cancel: cancel}, in.Offset, in.Whence)
+		off, errno := fs.Lseek(ctx, in.Offset, in.Whence)
 		out.Offset = off
 		return errnoToStatus(errno)
 	}
-
-	if in.Whence == _SEEK_DATA || in.Whence == _SEEK_HOLE {
+	var attr fuse.AttrOut
+	if s := b.getattr(ctx, n, nil, &attr); s != 0 {
+		return errnoToStatus(s)
+	}
+	if in.Whence == _SEEK_DATA {
+		if in.Offset >= attr.Size {
+			return errnoToStatus(syscall.ENXIO)
+		}
 		out.Offset = in.Offset
 		return fuse.OK
 	}
 
+	if in.Whence == _SEEK_HOLE {
+		if in.Offset > attr.Size {
+			return errnoToStatus(syscall.ENXIO)
+		}
+		out.Offset = attr.Size
+		return fuse.OK
+	}
+
 	return fuse.ENOTSUP
 }
