@@ -18,11 +18,6 @@ import (
 	"github.com/hanwen/go-fuse/v2/fuse"
 )
 
-type parentData struct {
-	name   string
-	parent *Inode
-}
-
 // StableAttr holds immutable attributes of a object in the filesystem.
 type StableAttr struct {
 	// Each Inode has a type, which does not change over the
@@ -32,8 +27,8 @@ type StableAttr struct {
 
 	// The inode number must be unique among the currently live
 	// objects in the file system. It is used to communicate to
-	// the kernel about this file object. The values uint64(-1),
-	// and 1 are reserved. When using Ino==0, a unique, sequential
+	// the kernel about this file object. The value uint64(-1)
+	// is reserved. When using Ino==0, a unique, sequential
 	// number is assigned (starting at 2^63 by default) on Inode creation.
 	Ino uint64
 
@@ -46,7 +41,7 @@ type StableAttr struct {
 
 // Reserved returns if the StableAttr is using reserved Inode numbers.
 func (i *StableAttr) Reserved() bool {
-	return i.Ino == 1 || i.Ino == ^uint64(0)
+	return i.Ino == ^uint64(0) // fuse.pollHackInode = ^uint64(0)
 }
 
 // Inode is a node in VFS tree.  Inodes are one-to-one mapped to
@@ -63,6 +58,11 @@ type Inode struct {
 	ops    InodeEmbedder
 	bridge *rawBridge
 
+	// The *Node ID* is an arbitrary uint64 identifier chosen by the FUSE library.
+	// It is used the identify *nodes* (files/directories/symlinks/...) in the
+	// communication between the FUSE library and the Linux kernel.
+	nodeId uint64
+
 	// Following data is mutable.
 
 	// file handles.
@@ -96,15 +96,15 @@ type Inode struct {
 
 	// Children of this Inode.
 	// When you change this, you MUST increment changeCounter.
-	children map[string]*Inode
+	children inodeChildren
 
 	// Parents of this Inode. Can be more than one due to hard links.
 	// When you change this, you MUST increment changeCounter.
-	parents map[parentData]struct{}
+	parents inodeParents
 }
 
 func (n *Inode) IsDir() bool {
-	return n.stableAttr.Mode&syscall.S_IFDIR != 0
+	return n.stableAttr.Mode&syscall.S_IFMT == syscall.S_IFDIR
 }
 
 func (n *Inode) embed() *Inode {
@@ -115,20 +115,20 @@ func (n *Inode) EmbeddedInode() *Inode {
 	return n
 }
 
-func initInode(n *Inode, ops InodeEmbedder, attr StableAttr, bridge *rawBridge, persistent bool) {
+func initInode(n *Inode, ops InodeEmbedder, attr StableAttr, bridge *rawBridge, persistent bool, nodeId uint64) {
 	n.ops = ops
 	n.stableAttr = attr
 	n.bridge = bridge
 	n.persistent = persistent
-	n.parents = make(map[parentData]struct{})
+	n.nodeId = nodeId
 	if attr.Mode == fuse.S_IFDIR {
-		n.children = make(map[string]*Inode)
+		n.children.init()
 	}
 }
 
 // Set node ID and mode in EntryOut
 func (n *Inode) setEntryOut(out *fuse.EntryOut) {
-	out.NodeId = n.stableAttr.Ino
+	out.NodeId = n.nodeId
 	out.Ino = n.stableAttr.Ino
 	out.Mode = (out.Attr.Mode & 07777) | n.stableAttr.Mode
 }
@@ -169,12 +169,8 @@ func modeStr(m uint32) string {
 func (n *Inode) String() string {
 	n.mu.Lock()
 	defer n.mu.Unlock()
-	var ss []string
-	for nm, ch := range n.children {
-		ss = append(ss, fmt.Sprintf("%q=i%d[%s]", nm, ch.stableAttr.Ino, modeStr(ch.stableAttr.Mode)))
-	}
 
-	return fmt.Sprintf("i%d (%s): %s", n.stableAttr.Ino, modeStr(n.stableAttr.Mode), strings.Join(ss, ","))
+	return fmt.Sprintf("i%d (%s): %s", n.stableAttr.Ino, modeStr(n.stableAttr.Mode), n.children.String())
 }
 
 // sortNodes rearranges inode group in consistent order.
@@ -259,7 +255,7 @@ func unlockNodes(ns ...*Inode) {
 func (n *Inode) Forgotten() bool {
 	n.mu.Lock()
 	defer n.mu.Unlock()
-	return n.lookupCount == 0 && len(n.parents) == 0 && !n.persistent
+	return n.lookupCount == 0 && n.parents.count() == 0 && !n.persistent
 }
 
 // Operations returns the object implementing the file system
@@ -277,34 +273,24 @@ func (n *Inode) Path(root *Inode) string {
 	var segments []string
 	p := n
 	for p != nil && p != root {
-		var pd parentData
-
 		// We don't try to take all locks at the same time, because
 		// the caller won't use the "path" string under lock anyway.
-		found := false
 		p.mu.Lock()
-		// Select an arbitrary parent
-		for pd = range p.parents {
-			found = true
-			break
-		}
+		// Get last known parent
+		pd := p.parents.get()
 		p.mu.Unlock()
-		if found == false {
+		if pd == nil {
 			p = nil
 			break
 		}
-		if pd.parent == nil {
-			break
-		}
-
 		segments = append(segments, pd.name)
 		p = pd.parent
 	}
 
 	if root != nil && root != p {
 		deletedPlaceholder := fmt.Sprintf(".go-fuse.%d/deleted", rand.Uint64())
-		n.bridge.logf("warning: Inode.Path: inode i%d is orphaned, replacing segment with %q",
-			n.stableAttr.Ino, deletedPlaceholder)
+		n.bridge.logf("warning: Inode.Path: n%d is orphaned, replacing segment with %q",
+			n.nodeId, deletedPlaceholder)
 		// NOSUBMIT - should replace rather than append?
 		segments = append(segments, deletedPlaceholder)
 	}
@@ -329,10 +315,13 @@ func (n *Inode) Path(root *Inode) string {
 // but it could be also valid if only iparent is locked and ichild was just
 // created and only one goroutine keeps referencing it.
 func (iparent *Inode) setEntry(name string, ichild *Inode) {
-	ichild.parents[parentData{name, iparent}] = struct{}{}
-	iparent.children[name] = ichild
-	ichild.changeCounter++
-	iparent.changeCounter++
+	if ichild.stableAttr.Mode == syscall.S_IFDIR {
+		// Directories cannot have more than one parent. Clear the map.
+		// This special-case is neccessary because ichild may still have a
+		// parent that was forgotten (i.e. removed from bridge.inoMap).
+		ichild.parents.clear()
+	}
+	iparent.children.set(iparent, name, ichild)
 }
 
 // NewPersistentInode returns an Inode whose lifetime is not in
@@ -360,8 +349,8 @@ func (n *Inode) ForgetPersistent() {
 // NewInode returns an inode for the given InodeEmbedder. The mode
 // should be standard mode argument (eg. S_IFDIR). The inode number in
 // id.Ino argument is used to implement hard-links.  If it is given,
-// and another node with the same ID is known, that will node will be
-// returned, and the passed-in `node` is ignored.
+// and another node with the same ID is known, the new inode may be
+// ignored, and the old one used instead.
 func (n *Inode) NewInode(ctx context.Context, node InodeEmbedder, id StableAttr) *Inode {
 	return n.newInode(ctx, node, id, false)
 }
@@ -380,8 +369,9 @@ func (n *Inode) removeRef(nlookup uint64, dropPersistence bool) (forgotten bool,
 	n.mu.Lock()
 	if nlookup > 0 && dropPersistence {
 		log.Panic("only one allowed")
+	} else if nlookup > n.lookupCount {
+		log.Panicf("n%d lookupCount underflow: lookupCount=%d, decrement=%d", n.nodeId, n.lookupCount, nlookup)
 	} else if nlookup > 0 {
-
 		n.lookupCount -= nlookup
 		n.changeCounter++
 	} else if dropPersistence && n.persistent {
@@ -389,14 +379,23 @@ func (n *Inode) removeRef(nlookup uint64, dropPersistence bool) (forgotten bool,
 		n.changeCounter++
 	}
 
+	n.bridge.mu.Lock()
+	if n.lookupCount == 0 {
+		forgotten = true
+		// Dropping the node from stableAttrs guarantees that no new references to this node are
+		// handed out to the kernel, hence we can also safely delete it from kernelNodeIds.
+		delete(n.bridge.stableAttrs, n.stableAttr)
+		delete(n.bridge.kernelNodeIds, n.nodeId)
+	}
+	n.bridge.mu.Unlock()
+
 retry:
 	for {
 		lockme = append(lockme[:0], n)
 		parents = parents[:0]
 		nChange := n.changeCounter
-		live = n.lookupCount > 0 || len(n.children) > 0 || n.persistent
-		forgotten = n.lookupCount == 0
-		for p := range n.parents {
+		live = n.lookupCount > 0 || n.children.len() > 0 || n.persistent
+		for _, p := range n.parents.all() {
 			parents = append(parents, p)
 			lockme = append(lockme, p.parent)
 		}
@@ -415,20 +414,17 @@ retry:
 		}
 
 		for _, p := range parents {
-			delete(p.parent.children, p.name)
-			p.parent.changeCounter++
+			if p.parent.children.get(p.name) != n {
+				// another node has replaced us already
+				continue
+			}
+			p.parent.children.del(p.parent, p.name)
 		}
-		n.parents = map[parentData]struct{}{}
-		n.changeCounter++
 
 		if n.lookupCount != 0 {
-			panic("lookupCount changed")
+			log.Panicf("n%d %p lookupCount changed: %d", n.nodeId, n, n.lookupCount)
 		}
 
-		n.bridge.mu.Lock()
-		delete(n.bridge.nodes, n.stableAttr.Ino)
-		n.bridge.mu.Unlock()
-
 		unlockNodes(lockme...)
 		break
 	}
@@ -446,7 +442,7 @@ retry:
 func (n *Inode) GetChild(name string) *Inode {
 	n.mu.Lock()
 	defer n.mu.Unlock()
-	return n.children[name]
+	return n.children.get(name)
 }
 
 // AddChild adds a child to this node. If overwrite is false, fail if
@@ -459,13 +455,10 @@ func (n *Inode) AddChild(name string, ch *Inode, overwrite bool) (success bool)
 retry:
 	for {
 		lockNode2(n, ch)
-		prev, ok := n.children[name]
+		prev := n.children.get(name)
 		parentCounter := n.changeCounter
-		if !ok {
-			n.children[name] = ch
-			ch.parents[parentData{name, n}] = struct{}{}
-			n.changeCounter++
-			ch.changeCounter++
+		if prev == nil {
+			n.children.set(n, name, ch)
 			unlockNode2(n, ch)
 			return true
 		}
@@ -481,11 +474,8 @@ retry:
 			continue retry
 		}
 
-		delete(prev.parents, parentData{name, n})
-		n.children[name] = ch
-		ch.parents[parentData{name, n}] = struct{}{}
-		n.changeCounter++
-		ch.changeCounter++
+		prev.parents.delete(parentData{name, n})
+		n.children.set(n, name, ch)
 		prev.changeCounter++
 		unlockNodes(lockme[:]...)
 
@@ -497,11 +487,16 @@ retry:
 func (n *Inode) Children() map[string]*Inode {
 	n.mu.Lock()
 	defer n.mu.Unlock()
-	r := make(map[string]*Inode, len(n.children))
-	for k, v := range n.children {
-		r[k] = v
-	}
-	return r
+	return n.children.toMap()
+}
+
+// childrenList returns the list of children of this directory Inode.
+// The result is guaranteed to be stable as long as the directory did
+// not change.
+func (n *Inode) childrenList() []childEntry {
+	n.mu.Lock()
+	defer n.mu.Unlock()
+	return n.children.list()
 }
 
 // Parents returns a parent of this Inode, or nil if this Inode is
@@ -509,10 +504,11 @@ func (n *Inode) Children() map[string]*Inode {
 func (n *Inode) Parent() (string, *Inode) {
 	n.mu.Lock()
 	defer n.mu.Unlock()
-	for k := range n.parents {
-		return k.name, k.parent
+	p := n.parents.get()
+	if p == nil {
+		return "", nil
 	}
-	return "", nil
+	return p.name, p.parent
 }
 
 // RmAllChildren recursively drops a tree, forgetting all persistent
@@ -546,7 +542,7 @@ retry:
 		lockme = append(lockme[:0], n)
 		nChange := n.changeCounter
 		for _, nm := range names {
-			ch := n.children[nm]
+			ch := n.children.get(nm)
 			if ch == nil {
 				n.mu.Unlock()
 				return false, true
@@ -556,23 +552,17 @@ retry:
 		n.mu.Unlock()
 
 		lockNodes(lockme...)
+
 		if n.changeCounter != nChange {
 			unlockNodes(lockme...)
-			// could avoid unlocking and relocking n here.
-			n.mu.Lock()
 			continue retry
 		}
 
 		for _, nm := range names {
-			ch := n.children[nm]
-			delete(n.children, nm)
-			delete(ch.parents, parentData{nm, n})
-
-			ch.changeCounter++
+			n.children.del(n, nm)
 		}
-		n.changeCounter++
 
-		live = n.lookupCount > 0 || len(n.children) > 0 || n.persistent
+		live = n.lookupCount > 0 || n.children.len() > 0 || n.persistent
 		unlockNodes(lockme...)
 
 		// removal successful
@@ -601,8 +591,8 @@ retry:
 		counter1 := n.changeCounter
 		counter2 := newParent.changeCounter
 
-		oldChild := n.children[old]
-		destChild := newParent.children[newName]
+		oldChild := n.children.get(old)
+		destChild := newParent.children.get(newName)
 		unlockNode2(n, newParent)
 
 		if destChild != nil && !overwrite {
@@ -616,27 +606,17 @@ retry:
 		}
 
 		if oldChild != nil {
-			delete(n.children, old)
-			delete(oldChild.parents, parentData{old, n})
-			n.changeCounter++
-			oldChild.changeCounter++
+			n.children.del(n, old)
 		}
 
 		if destChild != nil {
 			// This can cause the child to be slated for
 			// removal; see below
-			delete(newParent.children, newName)
-			delete(destChild.parents, parentData{newName, newParent})
-			destChild.changeCounter++
-			newParent.changeCounter++
+			newParent.children.del(newParent, newName)
 		}
 
 		if oldChild != nil {
-			newParent.children[newName] = oldChild
-			newParent.changeCounter++
-
-			oldChild.parents[parentData{newName, newParent}] = struct{}{}
-			oldChild.changeCounter++
+			newParent.children.set(newParent, newName, oldChild)
 		}
 
 		unlockNodes(n, newParent, oldChild, destChild)
@@ -658,8 +638,8 @@ retry:
 		counter1 := oldParent.changeCounter
 		counter2 := newParent.changeCounter
 
-		oldChild := oldParent.children[oldName]
-		destChild := newParent.children[newName]
+		oldChild := oldParent.children.get(oldName)
+		destChild := newParent.children.get(newName)
 		unlockNode2(oldParent, newParent)
 
 		if destChild == oldChild {
@@ -674,34 +654,20 @@ retry:
 
 		// Detach
 		if oldChild != nil {
-			delete(oldParent.children, oldName)
-			delete(oldChild.parents, parentData{oldName, oldParent})
-			oldParent.changeCounter++
-			oldChild.changeCounter++
+			oldParent.children.del(oldParent, oldName)
 		}
 
 		if destChild != nil {
-			delete(newParent.children, newName)
-			delete(destChild.parents, parentData{newName, newParent})
-			destChild.changeCounter++
-			newParent.changeCounter++
+			newParent.children.del(newParent, newName)
 		}
 
 		// Attach
 		if oldChild != nil {
-			newParent.children[newName] = oldChild
-			newParent.changeCounter++
-
-			oldChild.parents[parentData{newName, newParent}] = struct{}{}
-			oldChild.changeCounter++
+			newParent.children.set(newParent, newName, oldChild)
 		}
 
 		if destChild != nil {
-			oldParent.children[oldName] = oldChild
-			oldParent.changeCounter++
-
-			destChild.parents[parentData{oldName, oldParent}] = struct{}{}
-			destChild.changeCounter++
+			oldParent.children.set(oldParent, oldName, destChild)
 		}
 		unlockNodes(oldParent, newParent, oldChild, destChild)
 		return
@@ -712,7 +678,7 @@ retry:
 // tuple should be invalidated. On next access, a LOOKUP operation
 // will be started.
 func (n *Inode) NotifyEntry(name string) syscall.Errno {
-	status := n.bridge.server.EntryNotify(n.stableAttr.Ino, name)
+	status := n.bridge.server.EntryNotify(n.nodeId, name)
 	return syscall.Errno(status)
 }
 
@@ -721,7 +687,7 @@ func (n *Inode) NotifyEntry(name string) syscall.Errno {
 // to NotifyEntry, but also sends an event to inotify watchers.
 func (n *Inode) NotifyDelete(name string, child *Inode) syscall.Errno {
 	// XXX arg ordering?
-	return syscall.Errno(n.bridge.server.DeleteNotify(n.stableAttr.Ino, child.stableAttr.Ino, name))
+	return syscall.Errno(n.bridge.server.DeleteNotify(n.nodeId, child.nodeId, name))
 
 }
 
@@ -729,16 +695,16 @@ func (n *Inode) NotifyDelete(name string, child *Inode) syscall.Errno {
 // inode should be flushed from buffers.
 func (n *Inode) NotifyContent(off, sz int64) syscall.Errno {
 	// XXX how does this work for directories?
-	return syscall.Errno(n.bridge.server.InodeNotify(n.stableAttr.Ino, off, sz))
+	return syscall.Errno(n.bridge.server.InodeNotify(n.nodeId, off, sz))
 }
 
 // WriteCache stores data in the kernel cache.
 func (n *Inode) WriteCache(offset int64, data []byte) syscall.Errno {
-	return syscall.Errno(n.bridge.server.InodeNotifyStoreCache(n.stableAttr.Ino, offset, data))
+	return syscall.Errno(n.bridge.server.InodeNotifyStoreCache(n.nodeId, offset, data))
 }
 
 // ReadCache reads data from the kernel cache.
 func (n *Inode) ReadCache(offset int64, dest []byte) (count int, errno syscall.Errno) {
-	c, s := n.bridge.server.InodeRetrieveCache(n.stableAttr.Ino, offset, dest)
+	c, s := n.bridge.server.InodeRetrieveCache(n.nodeId, offset, dest)
 	return c, syscall.Errno(s)
 }
